#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <unistd.h>
#include <vector>
#include <fcntl.h>
#include <math.h>

#define MAXITER 20
#define MAXBOUNCE 10
#define BEAMSPEED 0.01
#define PROJDIST 0.7
#define EYEDIST  6
#define MAXTRACE ((EYEDIST+6)/BEAMSPEED)

#define XDIM (512/16)
#define YDIM (384/16)

#include <linux/fb.h>
static unsigned char Pal[768];
static class Video
{
    int fd;
    unsigned char *p;
    unsigned xres;
public:
    Video() : xres(1504)
    {
        fd = open("/dev/fb0", O_RDWR);
        
        struct fb_cmap cmap;
        cmap.start=0;
        cmap.len  =256;
        cmap.red  =(__u16 *)calloc(256, 2);
        cmap.green=(__u16 *)calloc(256, 2);
        cmap.blue =(__u16 *)calloc(256, 2);
        cmap.transp=NULL;

        ioctl(fd, FBIOGETCMAP, &cmap);
        for(int a=0; a<120; ++a)
        {
            double mag = a/120.0;
            
            int r = int(mag*511);
            int g = int(mag*511)-256;
            int b = int(mag*2400)-(2400-256);
            
            if(r>255)r=255;if(r<0)r=0;
            if(g>255)g=255;if(g<0)g=0;
            if(b>255)b=255;if(b<0)b=0;
            
            cmap.red  [a+16] = r * 65535 / 255;
            cmap.green[a+16] = g * 65535 / 255;
            cmap.blue [a+16] = b * 65535 / 255;
            
            cmap.red  [a+16+120] = b * 65535 / 255;
            cmap.green[a+16+120] = g * 65535 / 255;
            cmap.blue [a+16+120] = r * 65535 / 255;
        }
        for(int a=0; a<256; ++a)
        {
        	Pal[a*3+0] = cmap.blue [a]/256;
        	Pal[a*3+1] = cmap.green[a]/256;
        	Pal[a*3+2] = cmap.red  [a]/256;
        }
        ioctl(fd, FBIOPUTCMAP, &cmap);

        free(cmap.red);
        free(cmap.green);
        free(cmap.blue);

        p = (unsigned char *)mmap(NULL, xres*1024, PROT_WRITE, MAP_SHARED, fd, 0);
        if((int)p==-1)
        {
            perror("mmap");
            throw;
        }
    }
    ~Video()
    {
        munmap(p, xres*1024);
    }
    void operator() (unsigned x, unsigned y, unsigned char c)
    {
        /*p[y*xres+x] = c;*/
        
        x+=(1280-XDIM*2)/4; y+=(1024-YDIM*2)/4;
        
        p[(y*2  )*xres+(x*2  )] = c;
        p[(y*2  )*xres+(x*2+1)] = c;
        p[(y*2+1)*xres+(x*2  )] = c;
        p[(y*2+1)*xres+(x*2+1)] = c;
    }
} PSet;


class Matrix
{
    double val[4][4];
public:
    Matrix()
    {
    }
    Matrix(double ax, double ay, double az)
    {
        InitRotate(ax, ay, az);
    }
    void InitRotate(double ax, double ay, double az)
    {
        double tmp[3][4][4];

        memset(&tmp, 0, sizeof(tmp));

        tmp[0][1][1]=tmp[0][2][2]=cos(ax); tmp[0][2][1]=-(tmp[0][1][2]=sin(ax));
        tmp[1][0][0]=tmp[1][2][2]=cos(ay); tmp[1][0][2]=-(tmp[1][2][0]=sin(ay));
        tmp[2][0][0]=tmp[2][1][1]=cos(az); tmp[2][1][0]=-(tmp[2][0][1]=sin(az));

        tmp[0][0][0] = 1;
        tmp[0][3][3] = 1;
        tmp[1][1][1] = 1;
        tmp[1][3][3] = 1;
        tmp[2][2][2] = 1;
        tmp[2][3][3] = 1;

        for(int i=0; i<4; i++)
        {
            double m2[4];

            for(int j=0; j<4; j++)
                m2[j] = tmp[1][i][0] * tmp[0][0][j] +
                        tmp[1][i][1] * tmp[0][1][j] +
                        tmp[1][i][2] * tmp[0][2][j] +
                        tmp[1][i][3] * tmp[0][3][j];

            for(int j=0; j<4; j++)
                val[i][j] = m2[0] * tmp[2][0][j] +
                            m2[1] * tmp[2][1][j] +
                            m2[2] * tmp[2][2][j] +
                            m2[3] * tmp[2][3][j];
        }
    }
    template<class T>
    void Rotate(T &p)
    {
        register double tx = p.x*val[0][0] + p.y*val[1][0] + p.z*val[2][0] + val[3][0];
        register double ty = p.x*val[0][1] + p.y*val[1][1] + p.z*val[2][1] + val[3][1];

        p.z = p.x*val[0][2] + p.y*val[1][2] + p.z*val[2][2] + val[3][2];
        p.y = ty;
        p.x = tx;
    }
};

class XYZ
{
public:
    double x,y,z;
    
    XYZ(double xx,double yy,double zz) : x(xx),y(yy),z(zz)
    {
    }
    
    const XYZ &operator +=(const XYZ &p) { x+=p.x; y+=p.y; z+=p.z; return *this; }
    const XYZ &operator -=(const XYZ &p) { x-=p.x; y-=p.y; z-=p.z; return *this; }
    const XYZ &operator /=(const double &p) { x/=p; y/=p; z/=p; return *this; }
    const XYZ &operator *=(const double &p) { x*=p; y*=p; z*=p; return *this; }
    const XYZ operator +(const XYZ &p) const { return XYZ(x+p.x, y+p.y, z+p.z); }
    const XYZ operator -(const XYZ &p) const { return XYZ(x-p.x, y-p.y, z-p.z); }
    const XYZ operator /(const double &p) const { return XYZ(x/p, y/p, z/p); }
    const XYZ operator *(const double &p) const { return XYZ(x*p, y*p, z*p); }
    
    const double dist(const XYZ &p)const 
    {
        XYZ d(x-p.x, y-p.y, z-p.z);
        return sqrt(d.x*d.x + d.y*d.y + d.z*d.z);
    }
    const double length() const
    {
        return sqrt(x*x + y*y + z*z);
    }
    
    inline void Rotate(Matrix &m)
    {
        m.Rotate(*this);
    }
};

static class Fract
{
    inline bool ismandel(double RP,double IP)
    {
        if(IP<-1.15 || IP>1.15
        || RP<-2.02 || RP>0.5)return false;
        
        register double r = RP, i = IP;
        for(register int c=0; ; )
        {
            register const double r2 = r*r;
            register const double i2 = i*i;
            
            if(r2+i2 >= 4)
            {
                /* Would never return back to <4. Not part of set. Exit. */
                return false;
            }
            
            i = r*i*2 + IP;
            r = r2-i2 + RP;
            
            if(++c >= MAXITER)
            {
                /* Tried enough. We are inside. */
                return true;
            }
        }
    }

    inline bool isjulia(double XP,double YP, double RP,double IP)
    {
        if(!ismandel(RP, IP))return false;
        if(XP<-2 || XP>2
        || YP<-2 || YP>2)return false;
        
        register double r = XP, i = YP;
        for(register int c=0; ; )
        {
            register const double r2 = r*r;
            register const double i2 = i*i;
            
            if(r2+i2 >= 4)
            {
                /* Would never return back to <4. Not part of set. Exit. */
                return false;
            }
            
            i = r*i*2 + IP;
            r = r2-i2 + RP;
            
            if(++c >= MAXITER)
            {
                /* Tried enough. We are inside. */
                return true;
            }
        }
    }
    
public:
    
    inline bool hit(double x,double y, double r,double i)
    {
        return isjulia(x,y,r,i);
    }
} Fract;

static const double randang()
{
    return drand48() * 2 * M_PI;
}

int main(void)
{
    double w = -0.1; /* the time axis position */
    
    vector<XYZ> Lights;
    
    Lights.push_back(XYZ(2,0,0));
    
    for(double ax=0; ax<M_PI*4; ax+=M_PI / 100)
    {
        Matrix eyematrix(ax-M_PI, sin(ax), - sin(ax/2));
        
        char fnBuf[64]; static unsigned imgind=0;
	    unsigned char tgaBuf[18] = {0,0,2,0,0,0,0, 0,0,0,0,0, XDIM&255, XDIM>>8, YDIM&255, YDIM>>8, '', ' '};
        sprintf(fnBuf, "tga/mand%04u.tga", ++imgind);
        FILE *fp = fopen(fnBuf, "wb");
        fwrite(tgaBuf, 1, 18, fp);

        for(unsigned y=0; y<YDIM; ++y)
        {
            const double ydisp = (y / (double)(YDIM-1)*2) - 1;
            
            #define PSET(x,y,c) do { PSet(x,y,c); fwrite(&Pal[c*3], 1, 3, fp); } while(0)
            
            for(unsigned x=0; x<XDIM; ++x)
            {
                const double xdisp = (x / (double)(XDIM-1)*2) - 1;
                
                XYZ proj(xdisp,ydisp,EYEDIST-PROJDIST);
                XYZ dir = proj * (BEAMSPEED / proj.length());
                
                dir.Rotate(eyematrix);
                XYZ beam(0,0, -EYEDIST);
                beam.Rotate(eyematrix);

                unsigned bounce = 0;
                for(unsigned step=0; ; ++step)
                {
                    for(unsigned l=0; l<Lights.size(); ++l)
                        if(beam.dist(Lights[l]) < 0.2)
                        {
                            /* Hit the light */
                            if(!bounce)
                                PSET(x,y, 15);
                            else
                            {
                                double bright = 120 * sqrt(bounce) / sqrt(MAXBOUNCE) - 1E-6;
                                int c = 16 + 120 - 1 - unsigned(bright);
                                PSET(x,y, c);
                            }
                            goto DonePix;
                        }
                    if(step >= MAXTRACE)
                    {
                    Enough:
                        if(!bounce)
                        {
                            /* sky */
                            int c = 255-int(beam.dist(Lights[0])*120 / EYEDIST);
                            if(c < 16+121)c = 16+121;
                            PSET(x,y, c);
                        }
                        else
                            PSET(x,y, 18);
                        break;
                    }
                    if(Fract.hit(beam.z, w, beam.x, beam.y))
                    {
                        ++bounce;
                        if(bounce >= MAXBOUNCE)
                            goto Enough;
                        
                        dir = Lights[0] - beam;
                        dir /= dir.length();
                        dir *= BEAMSPEED;
                    }
                    beam += dir;
                }
DonePix:        ;
            }
        }
        
        fclose(fp);
    }
    return 0;
}
